#include <cstdlib>
#include <cmath>
#include "nv_math.h"
#include "Camera.h"

Camera::Camera() : up( vec3() ), position( vec3() ), center( vec3() ), perspectiveMatrix( NULL )
{}

Camera::~Camera()
{
	if ( perspectiveMatrix )
		delete perspectiveMatrix;
}

void Camera::setParameters( double posX, double posY, double posZ, 
			    double centerX, double centerY, double centerZ, 
			    double upX, double upY, double upZ, 
			    double fov )
{
	this->fov = fov * nv_to_rad;

	up.x = upX;
	up.y = upY;
	up.z = upZ;

	position.x = posX;
	position.y = posY;
	position.z = posZ;

	center.x = centerX;
	center.y = centerY;
	center.z = centerZ;

	cerr << "fov: " << this->fov << " (in deg=" << fov << ")" << endl;
	cerr << "up: [" << up.x << "," << up.y << "," << up.z << "]" << endl;
	cerr << "position: [" << position.x << "," << position.y << "," << position.z << "]" << endl;
	cerr << "center: [" << center.x << "," << center.y << "," << center.z << "]" << endl;
}

void Camera::getEyeBasis( vec3 &u, vec3 &v, vec3 &w )
{
	w = position - center;
	w.normalize();

	cross( u, up, w );
	u.normalize();

	cross( v, w, u );
}

mat3 Camera::eyeTransform()
{
	vec3 u, v, w;
	getEyeBasis( u, v, w );

	return mat3( u.x, u.y, u.z, v.x, v.y, v.z, w.x, w.y, w.z );
}

mat4 Camera::getPerspectiveMatrix( int width, int height, double near, double far )
{
	mat4 perspectiveMatrix;
	double f = cos( fov / 2 ) / sin( fov / 2 );

	perspectiveMatrix.a00 = f * ( (double)height / (double)width );
	perspectiveMatrix.a01 = 0.0;
	perspectiveMatrix.a02 = 0.0;
	perspectiveMatrix.a03 = 0.0;

	perspectiveMatrix.a10 = 0.0;
	perspectiveMatrix.a11 = f;
	perspectiveMatrix.a12 = 0.0;
	perspectiveMatrix.a13 = 0.0;

	perspectiveMatrix.a20 = 0.0;
	perspectiveMatrix.a21 = 0.0;
	perspectiveMatrix.a22 = ( far + near ) / ( near - far );
	perspectiveMatrix.a23 = ( 2 * far * near ) / ( near - far );

	perspectiveMatrix.a30 = 0.0;
	perspectiveMatrix.a31 = 0.0;
	perspectiveMatrix.a32 = -1.0;
	perspectiveMatrix.a33 = 0.0;

	return perspectiveMatrix;
}

// see raytracing nuts and bolts lecture notes
Ray Camera::getRay( unsigned int row, unsigned int col, unsigned int width, unsigned int height )
{
	double alpha, beta, aspect = (double)width / (double) height;
	alpha = tan( ( fov * aspect ) / 2.0 ) * ( ( col - width / 2.0 ) / ( width / 2.0 ) );
	beta = tan( fov / 2.0 ) * ( ( row - height / 2.0 ) / ( height / 2.0 ) );

	/*
#ifdef DEBUG
	cerr << "Camera::getRay alpha=" << alpha << ", beta=" << beta << endl;
	cerr << endl;
#endif
	*/

	vec3 u, v, w, nonEyeComponent;
	getEyeBasis( u, v, w );
	nonEyeComponent = alpha * u + beta * v - w;
	nonEyeComponent.normalize();

	return Ray( position, nonEyeComponent );
}

