#include <iostream>
#include <cstdlib>
#include "nv_math.h"
#include "nv_algebra.h"
#include "Light.h"
#include "Scene.h"
#include "Sphere.h"
#include "Triangle.h"
#include "Cone.h"
#include "TGAImage.h"


using namespace std;

static const int BUF_SIZE = 300;

// state variables used in the parsing
mat4 transformMat;
vector<mat4> transformStack;
vec3 diffuse;
vec3 specular;
vec3 emissive;
vec3 ambient;
double shininess;
double reflective;
double refractive;
double refractiveIndex;
double constantAttenuation;
double linearAttenuation;
double quadraticAttenuation;

Scene::Scene() : maxRecursionDepth( 10 ), outputFile( "out.tga" ), riggedRays( NULL ) {}


static int fstreamFailErr( fstream &fs, int charLimit = 100 )
{
      	if ( ( fs.fail() && !fs.eof() ) || fs.gcount() >= charLimit )
		return 1;
                                                                       
       	return fs.eof();
}


void fresnel( double etai, double etat, double cosi, double cost, double *Kr)
{
	double Rp = ((etat * cosi) - (etai * cost)) / ((etat * cosi ) + (etai * cost));
	double Rs = ((etai * cosi) - (etat * cost)) / ((etai * cosi ) + (etat * cost));
	
	*Kr = (Rp*Rp + Rs*Rs) * 0.5;
}

void Scene::doSize( vector<char*> tokens )
{
	width = atoi( tokens[ 0 ] );
	height = atoi( tokens[ 1 ] );

	pixels.clear();
	pixels.reserve( width );

	for ( int i = 0; i < height; i++ ) {
		vector<vec3> row;
		row.reserve( height );
		for ( int j = 0; j < width; j++ )
			row.push_back( vec3( 0.0, 0.0, 0.0 ) );

		pixels.push_back( row );
	}
}

void Scene::doMaxDepth( vector<char*> tokens )
{
	maxRecursionDepth = atoi( tokens[ 0 ] );
}

void Scene::doOutput( vector<char*> tokens )
{
	outputFile = tokens[ 0 ];
}

void Scene::doCamera( vector<char*> tokens )
{
	camera.setParameters( atof( tokens[ 0 ] ), atof( tokens[ 1 ] ), atof( tokens[ 2 ] ),
			      atof( tokens[ 3 ] ), atof( tokens[ 4 ] ), atof( tokens[ 5 ] ),
			      atof( tokens[ 6 ] ), atof( tokens[ 7 ] ), atof( tokens[ 8 ] ),
			      atof( tokens[ 9 ] )
			    );
}

void Scene::doSphere( vector<char*> tokens )
{
	double x, y, z, radius;
	x = atof( tokens[ 0 ] );
	y = atof( tokens[ 1 ] );
	z = atof( tokens[ 2 ] );
	radius = atof( tokens[ 3 ] );

	Sphere *s = new Sphere( vec3( x, y, z ), radius, transformMat, ambient, lights, 
				diffuse, specular, emissive, shininess, reflective, refractive );
	objects.push_back( s );
}
void Scene::doCone( vector<char*> tokens )
{
	double x, y, z, radius, height;
	x = atof( tokens[ 0 ] );
	y = atof( tokens[ 1 ] );
	z = atof( tokens[ 2 ] );
	radius = atof( tokens[ 3 ] );
	height = atof( tokens[ 4 ] );
	
	Cone *c = new Cone( vec3( x, y, z ), radius, height, transformMat, ambient, lights, diffuse, specular, emissive, shininess, reflective, refractive);
	objects.push_back( c );
}

void Scene::doVertex( vector<char*> tokens )
{
	double x, y, z;
	x = atof( tokens[ 0 ] );
	y = atof( tokens[ 1 ] );
	z = atof( tokens[ 2 ] );

	allVertices.push_back( Vertex( vec3( x, y, z ), diffuse, specular, emissive, shininess ) );
}

void Scene::doVertexNormal( vector<char*> tokens )
{
	int x, y, z, nx, ny, nz;
	x = atof( tokens[ 0 ] );
	y = atof( tokens[ 1 ] );
	z = atof( tokens[ 2 ] );
	nx = atof( tokens[ 3 ] );
	ny = atof( tokens[ 4 ] );
	nz = atof( tokens[ 5 ] );

	normals.push_back( vec4( nx, ny, nz, 1.0 ) );
	allVertexNormals.push_back( Vertex( vec3( x, y, z ), diffuse, specular, emissive, shininess, normals.back() ) );
}

void Scene::doTriangle( vector<char*> tokens )
{
	if ( tokens.size() != 3 )
	{
		cerr << "Invalid number of tokens passed to Scene::doTriangle:" << endl << "[";
		for ( unsigned int i = 0; i < tokens.size(); i++ )
			cerr << tokens[ i ] << " ";
		cerr << "]" << endl;

		exit( -1 );
	}

	int index1, index2, index3;
	index1 = atof( tokens[ 0 ] );
	index2 = atof( tokens[ 1 ] );
	index3 = atof( tokens[ 2 ] );

	if ( (unsigned int)index1 >= allVertices.size() || 
	     (unsigned int)index2 >= allVertices.size() || 
	     (unsigned int)index3 >= allVertices.size() ) 
	{
		cerr << "Invalid triangle vertices: " << index1 << ", " << index2 << ", " << index3 << endl;
		return;
	}

	Vertex v1 = allVertices[ index1 ];
	Vertex v2 = allVertices[ index2 ];
	Vertex v3 = allVertices[ index3 ];

	vec3 v1c = v1.coords;
	vec3 v2c = v2.coords;
	vec3 v3c = v3.coords;

	Triangle *t = new Triangle( v1, v2, v3, transformMat, ambient, lights,
		 			diffuse, specular, emissive, shininess, reflective, refractive );

	objects.push_back( t );
}

void Scene::doTriangleNormal( vector<char*> tokens )
{
	doTriangle( tokens ); // good code writes itself =P
}

void doTranslate( vector<char*> tokens )
{
	mat4 tranMat;
	tranMat.a00 = 1.0;
	tranMat.a01 = 0.0;
	tranMat.a02 = 0.0;
	tranMat.a03 = atof( tokens[ 0 ] );
	tranMat.a10 = 0.0;
	tranMat.a11 = 1.0;
	tranMat.a12 = 0.0;
	tranMat.a13 = atof( tokens[ 1 ] );
	tranMat.a20 = 0.0;
	tranMat.a21 = 0.0;
	tranMat.a22 = 1.0;
	tranMat.a23 = atof( tokens[ 2 ] );
	tranMat.a30 = 0.0;
	tranMat.a31 = 0.0;
	tranMat.a32 = 0.0;
	tranMat.a33 = 1.0;

	transformMat = transformMat * tranMat; 
}

void doRotate( vector<char*> tokens )
{
	mat3 rot;
	double x, y, z, angle;
	x = atof( tokens[ 0 ] );
	y = atof( tokens[ 1 ] );
	z = atof( tokens[ 2 ] );
	angle = atof( tokens[ 3 ] );

	vec3 v( x, y, z );
	v.normalize();

	rot.set_rot( angle * nv_to_rad, v );

	mat4 rot4;

	rot4.a00 = rot.a00;
	rot4.a01 = rot.a01;
	rot4.a02 = rot.a02;
	rot4.a03 = 0.0;

	rot4.a10 = rot.a10;
	rot4.a11 = rot.a11;
	rot4.a12 = rot.a12;
	rot4.a13 = 0.0;

	rot4.a20 = rot.a20;
	rot4.a21 = rot.a21;
	rot4.a22 = rot.a22;
	rot4.a23 = 0.0;

	rot4.a30 = 0.0;
	rot4.a31 = 0.0;
	rot4.a32 = 0.0;
	rot4.a33 = 1.0;

	transformMat = transformMat * rot4;
}

void doScale( vector<char*> tokens )
{
	double x, y, z;
	x = atof( tokens[ 0 ] );
	y = atof( tokens[ 1 ] );
	z = atof( tokens[ 2 ] );

	mat4 scaleMat;
	scaleMat.a00 = x;
	scaleMat.a01 = 0.0;
	scaleMat.a02 = 0.0;
	scaleMat.a03 = 0.0;

	scaleMat.a10 = 0.0;
	scaleMat.a11 = y;
	scaleMat.a12 = 0.0;
	scaleMat.a13 = 0.0;

	scaleMat.a20 = 0.0;
	scaleMat.a21 = 0.0;
	scaleMat.a22 = z;
	scaleMat.a23 = 0.0;

	scaleMat.a30 = 0.0;
	scaleMat.a31 = 0.0;
	scaleMat.a32 = 0.0;
	scaleMat.a33 = 1.0;

	transformMat = transformMat * scaleMat;
}

void doPush( vector<char*> tokens )
{
	transformStack.push_back( transformMat );
}

void doPop( vector<char*> tokens )
{
	transformMat = transformStack.back();
	transformStack.pop_back();
}

void Scene::doDirectionalLight( vector<char*> tokens )
{
	double x, y, z, r, g, b;
	x = atof( tokens[ 0 ] );
	y = atof( tokens[ 1 ] );
	z = atof( tokens[ 2 ] );
	r = atof( tokens[ 3 ] );
	g = atof( tokens[ 4 ] );
	b = atof( tokens[ 5 ] );

	Light dirLight = Light( r, g, b, x, y, z, Light::directional,
					      constantAttenuation,
					      linearAttenuation,
					      quadraticAttenuation
					    );

	lights.push_back( dirLight );
}

void Scene::doPointLight( vector<char*> tokens )
{
	double x, y, z, r, g, b;
	x = atof( tokens[ 0 ] );
	y = atof( tokens[ 1 ] );
	z = atof( tokens[ 2 ] );
	r = atof( tokens[ 3 ] );
	g = atof( tokens[ 4 ] );
	b = atof( tokens[ 5 ] );

	Light posLight = Light( r, g, b, x, y, z, Light::point,
					      constantAttenuation,
					      linearAttenuation,
					      quadraticAttenuation
					    );

	lights.push_back( posLight );
}

void doAttenuation( vector<char*> tokens )
{
	constantAttenuation = atof( tokens[ 0 ] );
	linearAttenuation = atof( tokens[ 1 ] );
	quadraticAttenuation = atof( tokens[ 2 ] );
}

void doAmbient( vector<char*> tokens )
{
	ambient[ 0 ] = atof( tokens[ 0 ] );
	ambient[ 1 ] = atof( tokens[ 1 ] );
	ambient[ 2 ] = atof( tokens[ 2 ] );
}

void doDiffuse( vector<char*> tokens )
{
	diffuse[ 0 ] = atof( tokens[ 0 ] );
	diffuse[ 1 ] = atof( tokens[ 1 ] );
	diffuse[ 2 ] = atof( tokens[ 2 ] );
}

void doSpecular( vector<char*> tokens )
{
	specular[ 0 ] = atof( tokens[ 0 ] );
	specular[ 1 ] = atof( tokens[ 1 ] );
	specular[ 2 ] = atof( tokens[ 2 ] );
}

void doShininess( vector<char*> tokens )
{
	shininess = atof( tokens[ 0 ] );
}

void doReflective( vector<char*> tokens )
{
	reflective = atof( tokens[ 0 ] );
}

void doRefractive( vector<char*> tokens )
{
	refractive = atof( tokens[ 0 ] );
}

void doRefractiveIndex( vector<char*> tokens )
{
	refractiveIndex = atof( tokens[ 0 ] );
}

void doEmissive( vector<char*> tokens )
{
	emissive[ 0 ] = atof( tokens[ 0 ] );
	emissive[ 1 ] = atof( tokens[ 1 ] );
	emissive[ 2 ] = atof( tokens[ 2 ] );
}

void Scene::initStateVars()
{
	transformMat.a00 = 1.0;
	transformMat.a01 = 0.0;
	transformMat.a02 = 0.0;
	transformMat.a03 = 0.0;
	transformMat.a10 = 0.0;
	transformMat.a11 = 1.0;
	transformMat.a12 = 0.0;
	transformMat.a13 = 0.0;
	transformMat.a20 = 0.0;
	transformMat.a21 = 0.0;
	transformMat.a22 = 1.0;
	transformMat.a23 = 0.0;
	transformMat.a30 = 0.0;
	transformMat.a31 = 0.0;
	transformMat.a32 = 0.0;
	transformMat.a33 = 1.0;

	transformStack.clear();

	constantAttenuation = 1.0;
	linearAttenuation = 0.0;
	quadraticAttenuation = 0.0;

	diffuse[ 0 ] = 0.5;
	diffuse[ 1 ] = 0.5;
	diffuse[ 2 ] = 0.5;

	specular[ 0 ] = 0.5;
	specular[ 1 ] = 0.5;
	specular[ 2 ] = 0.5;

	emissive[ 0 ] = 0.0;
	emissive[ 1 ] = 0.0;
	emissive[ 2 ] = 0.0;

	ambient[ 0 ] = 0.2;
	ambient[ 1 ] = 0.2;
	ambient[ 2 ] = 0.2;

	shininess = 0.3;
	
	reflective = 0.0;
	refractive = 0.0;
	refractiveIndex = 1.0;
}

bool Scene::parse( fstream &descriptor )
{
	initStateVars();

	int lineNumber = 0;
	char buf[ BUF_SIZE ];
	while ( descriptor.good() )
	{
		descriptor.getline( buf, BUF_SIZE );
		if ( fstreamFailErr( descriptor, BUF_SIZE ) )
			break;
		lineNumber++;

		char *command = strtok( buf, " \n" );
		if ( !command )
			continue; // empty line

		char *token;
		vector<char*> tokens;
		while ( ( token = strtok( NULL, " \n" ) ) )
			tokens.push_back( token );

		if ( strcmp( command, "size" ) == 0 && tokens.size() == 2 )
			doSize( tokens );
		else if ( strcmp( command, "maxdepth" ) == 0 && tokens.size() == 1 )
			doMaxDepth( tokens );
		else if ( strcmp( command, "output" ) == 0 && tokens.size() == 1 )
			doOutput( tokens );
		else if ( strcmp( command, "camera" ) == 0 && tokens.size() == 10 )
			doCamera( tokens );
		else if ( strcmp( command, "sphere" ) == 0 && tokens.size() == 4 )
			doSphere( tokens );
		else if ( strcmp( command, "cone" ) == 0 && tokens.size() == 5 )
			doCone( tokens );
		else if ( strcmp( command, "maxverts" ) == 0 && tokens.size() == 1 )
			continue;
		else if ( strcmp( command, "maxvertnorms" ) == 0 && tokens.size() == 1 )
			continue;
		else if ( strcmp( command, "vertex" ) == 0 && tokens.size() == 3 )
			doVertex( tokens );
		else if ( strcmp( command, "vertexnormal" ) == 0 && tokens.size() == 6 )
			doVertexNormal( tokens );
		else if ( strcmp( command, "tri" ) == 0 && tokens.size() == 3 )
			doTriangle( tokens );
		else if ( strcmp( command, "trinormal" ) == 0 && tokens.size() == 3 )
			doTriangleNormal( tokens );
		else if ( strcmp( command, "translate" ) == 0 && tokens.size() == 3 )
			doTranslate( tokens );
		else if ( strcmp( command, "rotate" ) == 0 && tokens.size() == 4 )
			doRotate( tokens );
		else if ( strcmp( command, "scale" ) == 0 && tokens.size() == 3 )
			doScale( tokens );
		else if ( strcmp( command, "pushTransform" ) == 0 && tokens.size() == 0 )
			doPush( tokens );
		else if ( strcmp( command, "popTransform" ) == 0 && tokens.size() == 0 )
			doPop( tokens );
		else if ( strcmp( command, "directional" ) == 0 && tokens.size() == 6 )
			doDirectionalLight( tokens );
		else if ( strcmp( command, "point" ) == 0 && tokens.size() == 6 )
			doPointLight( tokens );
		else if ( strcmp( command, "attenuation" ) == 0 && tokens.size() == 3 )
			doAttenuation( tokens );
		else if ( strcmp( command, "ambient" ) == 0 && tokens.size() == 3 )
			doAmbient( tokens );
		else if ( strcmp( command, "diffuse" ) == 0 && tokens.size() == 3 )
			doDiffuse( tokens );
		else if ( strcmp( command, "specular" ) == 0 && tokens.size() == 3 )
			doSpecular( tokens );
		else if ( strcmp( command, "shininess" ) == 0 && tokens.size() == 1 )
			doShininess( tokens );
		else if ( strcmp( command, "reflective" ) == 0 && tokens.size() == 1 )
			doReflective( tokens );
		else if ( strcmp( command, "refractive" ) == 0 && tokens.size() == 1 )
			doRefractive( tokens );
		else if ( strcmp( command, "refractiveIndex" ) == 0 && tokens.size() == 1 )
			doRefractiveIndex( tokens );
		else if ( strcmp( command, "emission" ) == 0 && tokens.size() == 3 )
			doEmissive( tokens );
		else if ( command && command[ 0 ] == '#' )
			continue;
		else if ( command ) {
			cerr << "Ignoring malformed/unrecognized command '" << command << "' on line " << lineNumber << endl;
			cerr << " tokens:" << endl;
			for ( unsigned int i = 0; i < tokens.size(); i++ )
				cerr << "\t" << tokens[ i ] << endl;
			cerr << endl;
			continue;
		}
		else {
			cerr << "Unable to parse line " << lineNumber << " of input file" << endl;
			continue;
		}
	}

	if ( !descriptor.eof() )
		return false;

	// insert checks here for stuff like no camera specified, etc. ?

	return true;
}


void Scene::renderImage()
{
	cerr << "Scene::renderImage" << endl << endl;

	// make one ray per pixel
	vector< vector<Ray> > rays;

	for ( unsigned int row = 0; row < pixels.size(); row++ )
	{
		vector<Ray> rayRow;
		for ( unsigned int col = 0; col < pixels[ row ].size(); col++ )
			rayRow.push_back( camera.getRay( row, col, pixels[ row ].size(), pixels.size() ) );

		rays.push_back( rayRow );
	}

	for ( unsigned int row = 0; row < rays.size(); row++ )
		for ( unsigned int col = 0; col < rays[ row ].size(); col++ )
			pixels[ row ][ col ] = recursiveTrace( rays[ row ][ col ], maxRecursionDepth );
}

//saves scene as a .tga image
void Scene::saveImage()
{
	int height = pixels.size();
	if ( height == 0 )
		return;

	int width = pixels[ 0 ].size();
	
	int imageSize = width * height * 4;
	unsigned char *data = (unsigned char *)malloc( sizeof( unsigned char ) * imageSize );

	int bytesPerRow = width * 4;
	double largest = 0.0;
	for ( int i = 0; i < height; i++ ) {
		for ( int j = 0; j < width; j++ ) {
			if ( pixels[ i ][ j ][ 0 ] > largest )
				largest = pixels[ i ][ j ][ 0 ];
			if ( pixels[ i ][ j ][ 1 ] > largest )
				largest = pixels[ i ][ j ][ 1 ];
			if ( pixels[ i ][ j ][ 2 ] > largest )
				largest = pixels[ i ][ j ][ 2 ];
		}
	}

	cerr << "largest pixel value: " << largest << endl;

	double conversionFactor = 255 / largest; // use to scale values from 0 to 255

	cerr << hex;
	for ( int dataIndex = 0; dataIndex < imageSize; dataIndex += 4 )
	{
		int row = dataIndex / bytesPerRow;
		int col = ( dataIndex % bytesPerRow ) / 4;

		data[ dataIndex ] = (unsigned char)( pixels[ row ][ col ][ 0 ] * conversionFactor );
		data[ dataIndex + 1 ] = (unsigned char)( pixels[ row ][ col ][ 1 ] * conversionFactor );
		data[ dataIndex + 2 ] = (unsigned char)( pixels[ row ][ col ][ 2 ] * conversionFactor );
		data[ dataIndex + 3 ] = 0xff;
	}

	TGAImage::write( outputFile, data, width, height );
	free( data );
}

// modifies passed reflectionDir and refractionDir vectors according to Snell's law
//   the passed incident ray must have hit an object
void snell (Ray incident, Ray *reflection, Ray *refraction)
{
	SceneObject *hitObject = incident.getHitObject();
	bool hitIsReflective = hitObject->getMaterial().getReflective() > 0.0;
	bool hitIsRefractive = hitObject->getMaterial().getRefractive() > 0.0;

	vec3 normal = incident.getNormalAtHit();
	vec3 incidentDir = incident.getPath();
	double cost1 = dot( incidentDir, normal ); // cosine of "theta1", the incident angle
	
	if ( hitIsReflective )
	{
		vec3 reflectionDir = incidentDir - 2 * cost1 * normal;
		*reflection = Ray( incident.valueAtHit(), reflectionDir );
	}

	// skip refraction if material not refractive
	if ( !hitIsRefractive )
		return;

	double n1n2 =  1.0/hitObject->getMaterial().getRefractiveIndex(); // assume from air to substance
	double cost2 = sqrt(1.0 - (n1n2 * n1n2) * (1.0 - cost1*cost1)); // cosine of "theta2", the refracted angle
	vec3 internalRefractionDir = n1n2 * incidentDir + (n1n2 * cost1 + cost2 ) * normal;

	*refraction = Ray( incident.valueAtHit(), internalRefractionDir );

	//cerr << "\tat hit2: " << incident.valueAtHit() << endl;

	// now travel along the refraction ray until exiting the other side of the object

	refraction->moveEyeAlongPath( 0.01 ); // move a bit away from the edge to handle numerical errors

	hitObject->intersect( *refraction );

	if ( !refraction->hitAnObject() )
	{

		
#ifdef LIGHT_DEBUG
		cerr << "!! BADNESS: refracted ray did NOT hit the other side of the refractive object!" << endl;
		cerr << "\tcost1: " << cost1 << endl;
		cerr << "\torigin: " << refraction->getEye() << endl;
		cerr << "\t\torigin norm: " << refraction->getEye().norm() << endl;
		cerr << "\tpath: " << refraction->getPath() << endl;
		cerr << "\tincidentHit: " << incident.valueAtHit() << endl;
		cerr << "\t\tincidentHit norm: " << incident.valueAtHit().norm() << endl;
		cerr << "\tinternalRefractionDir" << internalRefractionDir << endl;
		cerr << endl;
#endif 

	//	*refraction = Ray( incident.valueAtHit(), reflectionDir );
		*refraction = Ray( incident.valueAtHit(), vec3( -0.01, -0.01, -0.01 ) );
		return;
	}

	// invert n1n2 because going the other way -- from different material back to air
	// WARNING: THIS IS AN ASSUMPTION !!!!!!!!!!!!!
	double n2n1 = 1.0 / n1n2;
	normal = refraction->getNormalAtHit();
	incidentDir = refraction->getPath();
	incidentDir.normalize();
	cost1 = dot( incidentDir, normal );
	cost2 = sqrt( 1.0 - n2n1*n2n1 * ( 1.0 - cost1*cost1 ) );

	vec3 outgoingRefractionDir = n2n1 * incidentDir + (n2n1 * cost1 + cost2) * normal;
	outgoingRefractionDir.normalize();
	*refraction = Ray( refraction->valueAtHit(), outgoingRefractionDir );
}

vec3 Scene::recursiveTrace( Ray ray, int remainingBounces )
{	
	if( !remainingBounces )
		return ray.hitAnObject() ? ray.colorAtHit(objects, ray.getEye()+ray.getPath())
				: vec3( 0.0, 0.0, 0.0 );

	for ( unsigned int objIndex = 0; objIndex < objects.size(); objIndex++ )
		objects[ objIndex ]->intersect( ray );

	if ( ray.hitAnObject() )
	{
		double reflective = ray.getHitObject()->getMaterial().getReflective();
		double refractive = ray.getHitObject()->getMaterial().getRefractive();
		vec3 colorAtHit = ray.colorAtHit( objects, camera.getPosition() );

		Ray reflection;
		Ray refraction;
		snell( ray, &reflection, &refraction );

		vec3 reflectionColor = reflective != 0.0 ? recursiveTrace( reflection, remainingBounces - 1 ) : vec3( 0.0, 0.0, 0.0 );
		vec3 refractionColor = refractive != 0.0 ? recursiveTrace( refraction, remainingBounces - 1 ) : vec3( 0.0, 0.0, 0.0 );

		return reflective * reflectionColor + refractive * refractionColor + ( 1.0 - reflective - refractive ) * colorAtHit;
		//return reflective * reflectionColor + ( 1.0 - reflective ) * colorAtHit;
	}

	return vec3( 0.0, 0.0, 0.0 );
}

