#include "SceneObject.h"

int SceneObject::nextId = 0;

SceneObject::SceneObject() {}

SceneObject::SceneObject( mat4 transformation, vec3 ambient, vector<Light> lights, vec3 matDiffuse,
		  vec3 matSpecular, vec3 matEmissive, double shininess, double reflective, double refractive ) :
	lights( lights ), material( Material( matDiffuse, matSpecular, matEmissive, shininess, reflective, refractive) ),
	transformation( transformation ), ambientLight( ambient ), id( ++nextId )
{
	invert( inverseTransformation, transformation );

	mat4 normalTransform = inverseTransformation;
	transpose( normalTransform );

	invert( normalRemapMat, normalTransform );
}

SceneObject::~SceneObject() {}


bool SceneObject::positionalLightBlocked( Light light, vec3 pos, vector<SceneObject*> objects )
{
	double tLight = 1.0; // by setup, ray hits light at t=1
	Ray rayToLight( pos, light.getCoords() - pos );
	rayToLight.moveEyeAlongPath( epsilon );

	bool blocked = false;
	for ( unsigned int i = 0; i < objects.size(); i++ )
	{
		objects[ i ]->intersect( rayToLight );
		if ( rayToLight.hitAnObject() && rayToLight.tAtHit() < tLight )
		{
			blocked = true;
			break;
		}
	}

	return blocked;
}

bool SceneObject::directionalLightBlocked( Light light, vec3 pos, vector<SceneObject*> objects )
{
	// directional light coords ARE the path, see Light constructor
	Ray lightRay( pos, light.getCoords() ); 
	lightRay.moveEyeAlongPath( epsilon );

	bool blocked = false;
	for ( unsigned int i = 0; i < objects.size(); i++ )
	{
		objects[ i ]->intersect( lightRay );
		if ( lightRay.hitAnObject() )
		{
			blocked = true;
			break;
		}
	}

	return blocked;
}

const double SceneObject::epsilon = 0.1;

vector<unsigned int> SceneObject::getVisibleLightIndices( vec3 pos, vector<SceneObject*> objects )
{
	vector<unsigned int> visibleIndices;
	for ( unsigned int i = 0; i < lights.size(); i++ )
	{
		Light currentLight = lights[ i ];

		bool blocked = false;
		if ( currentLight.isPointLight() )
			blocked = positionalLightBlocked( currentLight, pos, objects );
		else if ( currentLight.isDirectionalLight() )
			blocked = directionalLightBlocked( currentLight, pos, objects );
		else
		{
			cerr << "Unrecognized light type in SceneObject::getVisibleLightIndices!" << endl;
			continue;
		}

		if ( blocked )
			continue;
		visibleIndices.push_back( i );
	}

	return visibleIndices;
}

vec3 SceneObject::getColorAt( Ray &ray, vector<SceneObject*> objects, vec3 viewerPos )
{

	vec3 color = ambientLight + material.getEmissive();
	vector<unsigned int> visibleLightIndices = 
			this->SceneObject::getVisibleLightIndices( ray.valueAtHit(), objects );

	for ( unsigned int i = 0; i < visibleLightIndices.size(); i++ )
	{
		Light *light = &lights[ visibleLightIndices[ i ] ];
		vec3 toLight;

		if ( light->isPointLight() )
			toLight = light->getCoords() - ray.valueAtHit();
		else if ( light->isDirectionalLight() )
			toLight = light->getCoords(); // see comment in light constructor
		else 
		{
			cerr << "Unimplemented light type in Sphere::getColorAt!" << endl;
			continue;
		}

		double lightDistance = toLight.norm();
		toLight.normalize();

		vec3 normal = ray.getNormalAtHit();

		// diffuse component
		double directionScale = MAX( dot( toLight, normal ), 0.0 );
		vec3 lightColor = light->valueAt( lightDistance );
		vec3 diffuseMat = material.getDiffuse();
		vec3 netColor( lightColor.x * diffuseMat.x, lightColor.y * diffuseMat.y, lightColor.z * diffuseMat.z );
		color += directionScale * netColor;

		if ( material.getShininess() >= 0 ) 
		{
			vec3 halfAngle = (toLight + viewerPos);
			halfAngle.normalize();

			color += pow( (double)dot( normal, halfAngle ), (double)material.getShininess() ) * material.getSpecular();
		}

	}

	return color;
}

