#ifndef SCENE_OBJECT_H_
#define SCENE_OBJECT_H_

#include <vector>

#include "Light.h"
#include "Material.h"
#include "Vertex.h"

#include <iostream>

#define MAX(x,y) ( (x) > (y) ? (x) : (y) )

using namespace std;

class Ray;

class SceneObject
{
private:
	static int nextId;
	const static double epsilon;

	// utility functions for lighting calculations
	bool positionalLightBlocked( Light light, vec3 pos, vector<SceneObject*> objects );
	bool directionalLightBlocked( Light light, vec3 pos, vector<SceneObject*> objects );

protected:
	// Vertex objects hold normal information internally
	vector<Vertex> vertices;  // stored in transformed coordinates as per "transformation"
	vector<Light> lights; // all lights acting on the vertex
	Material material;
	mat4 transformation;
	mat4 inverseTransformation; // pre-calculated inverse
	mat4 normalRemapMat; // matrix for taking normals from object space to world space
	vec3 ambientLight; // r,g and b values
	double Kr;
	int id;

public:
	SceneObject();
	SceneObject( mat4 transformation, vec3 ambient, vector<Light> lights, vec3 matDiffuse,
		  vec3 matSpecular, vec3 matEmissive, double shininess, double reflective, double refractive );
	virtual ~SceneObject();

	inline void setTransformation( mat4 &m ) { transformation = m; }
	inline void setMaterial( vec3 diffuse, vec3 specular, vec3 emissive, double shininess )
	{
		material = Material( diffuse, specular, emissive, shininess );
	}
	inline Material getMaterial() { return material; }
	inline double getKr() {return Kr; } 
	inline int getId() { return id; }
	inline void setAmbient( vec3 ambient ) { ambientLight = ambient; }

	// updates the ray if the calculated "t" value is valid and less than current t value,
	//   sets normalAtHit of ray if necessary
	inline virtual void intersect( Ray &ray ) { cerr << "SceneObject::intersect = BAD" << endl; }

	
	/* does NOT ignore self, so if ray.valueAtHit() is negative because of no hit, object blocks itself
	 *    handles moving the provided position along direction to light to handle numerical inaccuracies
	 *
	 * pos : the position of the hit
	 */
	vector<unsigned int> getVisibleLightIndices( vec3 pos, vector<SceneObject*> objects );
	vec3 getColorAt( Ray &ray, vector<SceneObject*> otherObjects, vec3 viewerPos );

};

#include "Ray.h"

#endif

